package com.sun.org.apache.xpath.internal.axes;

import com.sun.org.apache.xpath.internal.XPathVisitor;
import com.sun.org.apache.xpath.internal.ExpressionOwner;
import java.util.Vector;
import com.sun.org.apache.xml.internal.dtm.DTM;
import java.io.IOException;
import java.io.ObjectInputStream;
import javax.xml.transform.TransformerException;
import com.sun.org.apache.xpath.internal.compiler.OpMap;
import com.sun.org.apache.xpath.internal.compiler.Compiler;
import com.sun.org.apache.xpath.internal.Expression;
import com.sun.org.apache.xml.internal.utils.WrappedRuntimeException;
import java.io.Serializable;
import com.sun.org.apache.xml.internal.dtm.DTMIterator;

public class UnionPathIterator extends LocPathIterator implements Cloneable, DTMIterator, Serializable, PathComponent
{
    static final long serialVersionUID = -3910351546843826781L;
    protected LocPathIterator[] m_exprs;
    protected DTMIterator[] m_iterators;
    
    public UnionPathIterator() {
        this.m_iterators = null;
        this.m_exprs = null;
    }
    
    @Override
    public void setRoot(int context, Object environment) {
        super.setRoot(context, environment);
        try {
            if (null != this.m_exprs) {
                int n = this.m_exprs.length;
                DTMIterator[] newIters = new DTMIterator[n];
                for (int i = 0; i < n; ++i) {
                    DTMIterator iter = this.m_exprs[i].asIterator(this.m_execContext, context);
                    (newIters[i] = iter).nextNode();
                }
                this.m_iterators = newIters;
            }
        }
        catch (Exception e) {
            throw new WrappedRuntimeException(e);
        }
    }
    
    public void addIterator(DTMIterator expr) {
        if (null == this.m_iterators) {
            (this.m_iterators = new DTMIterator[1])[0] = expr;
        }
        else {
            DTMIterator[] exprs = this.m_iterators;
            int len = this.m_iterators.length;
            System.arraycopy(exprs, 0, this.m_iterators = new DTMIterator[len + 1], 0, len);
            this.m_iterators[len] = expr;
        }
        expr.nextNode();
        if ((expr instanceof Expression)) {
            ((Expression)expr).exprSetParent(this);
        }
    }
    
    @Override
    public void detach() {
        if (this.m_allowDetach && null != this.m_iterators) {
            for (int n = this.m_iterators.length, i = 0; i < n; ++i) {
                this.m_iterators[i].detach();
            }
            this.m_iterators = null;
        }
    }
    
    public UnionPathIterator(Compiler compiler, int opPos) throws TransformerException {
        opPos = OpMap.getFirstChildPos(opPos);
        this.loadLocationPaths(compiler, opPos, 0);
    }
    
    public static LocPathIterator createUnionIterator(Compiler compiler, int opPos) throws TransformerException {
        UnionPathIterator upi = new UnionPathIterator(compiler, opPos);
        int nPaths = upi.m_exprs.length;
        boolean isAllChildIterators = true;
        int i = 0;
        while (i < nPaths) {
            LocPathIterator lpi = upi.m_exprs[i];
            if (lpi.getAxis() != 3) {
                isAllChildIterators = false;
                break;
            }
            if (HasPositionalPredChecker.check(lpi)) {
                isAllChildIterators = false;
                break;
            }
            ++i;
        }
        if (isAllChildIterators) {
            UnionChildIterator uci = new UnionChildIterator();
            for (int j = 0; j < nPaths; ++j) {
                PredicatedNodeTest lpi2 = upi.m_exprs[j];
                uci.addNodeTest(lpi2);
            }
            return uci;
        }
        return upi;
    }
    
    @Override
    public int getAnalysisBits() {
        int bits = 0;
        if (this.m_exprs != null) {
            for (int n = this.m_exprs.length, i = 0; i < n; ++i) {
                int bit = this.m_exprs[i].getAnalysisBits();
                bits |= bit;
            }
        }
        return bits;
    }
    
    private void readObject(ObjectInputStream stream) throws IOException, TransformerException {
        try {
            stream.defaultReadObject();
            this.m_clones = new IteratorPool(this);
        }
        catch (ClassNotFoundException cnfe) {
            throw new TransformerException(cnfe);
        }
    }
    
    @Override
    public Object clone() throws CloneNotSupportedException {
        UnionPathIterator clone = (UnionPathIterator)super.clone();
        if (this.m_iterators != null) {
            int n = this.m_iterators.length;
            clone.m_iterators = new DTMIterator[n];
            for (int i = 0; i < n; ++i) {
                clone.m_iterators[i] = (DTMIterator)this.m_iterators[i].clone();
            }
            return clone;
        }
        return clone;
    }
    
    protected LocPathIterator createDTMIterator(Compiler compiler, int opPos) throws TransformerException {
        LocPathIterator lpi = (LocPathIterator)WalkerFactory.newDTMIterator(compiler, opPos, compiler.getLocationPathDepth() <= 0);
        return lpi;
    }
    
    protected void loadLocationPaths(Compiler compiler, int opPos, int count) throws TransformerException {
        int steptype = compiler.getOp(opPos);
        if (steptype == 28) {
            this.loadLocationPaths(compiler, compiler.getNextOpPos(opPos), count + 1);
            (this.m_exprs[count] = this.createDTMIterator(compiler, opPos)).exprSetParent(this);
        }
        else {
            switch (steptype) {
                case 22:
                case 23:
                case 24:
                case 25: {
                    this.loadLocationPaths(compiler, compiler.getNextOpPos(opPos), count + 1);
                    WalkingIterator iter = new WalkingIterator(compiler.getNamespaceContext());
                    iter.exprSetParent(this);
                    if (compiler.getLocationPathDepth() <= 0) {
                        iter.setIsTopLevel(true);
                    }
                    (iter.m_firstWalker = new FilterExprWalker(iter)).init(compiler, opPos, steptype);
                    this.m_exprs[count] = iter;
                    break;
                }
                default: {
                    this.m_exprs = new LocPathIterator[count];
                    break;
                }
            }
        }
    }
    
    @Override
    public int nextNode() {
        if (this.m_foundLast) {
            return -1;
        }
        int earliestNode = -1;
        if (null != this.m_iterators) {
            int n = this.m_iterators.length;
            int iteratorUsed = -1;
            for (int i = 0; i < n; ++i) {
                int node = this.m_iterators[i].getCurrentNode();
                if (-1 != node) {
                    if (-1 == earliestNode) {
                        iteratorUsed = i;
                        earliestNode = node;
                    }
                    else if (node == earliestNode) {
                        this.m_iterators[i].nextNode();
                    }
                    else {
                        DTM dtm = this.getDTM(node);
                        if (dtm.isNodeAfter(node, earliestNode)) {
                            iteratorUsed = i;
                            earliestNode = node;
                        }
                    }
                }
            }
            if (-1 != earliestNode) {
                this.m_iterators[iteratorUsed].nextNode();
                this.incrementCurrentPos();
            }
            else {
                this.m_foundLast = true;
            }
        }
        return this.m_lastFetched = earliestNode;
    }
    
    @Override
    public void fixupVariables(Vector vars, int globalsSize) {
        for (int i = 0; i < this.m_exprs.length; ++i) {
            this.m_exprs[i].fixupVariables(vars, globalsSize);
        }
    }
    
    @Override
    public int getAxis() {
        return -1;
    }
    
    @Override
    public void callVisitors(ExpressionOwner owner, XPathVisitor visitor) {
        if (visitor.visitUnionPath(owner, this) && null != this.m_exprs) {
            for (int n = this.m_exprs.length, i = 0; i < n; ++i) {
                this.m_exprs[i].callVisitors(new UnionPathIterator.iterOwner(i), visitor);
            }
        }
    }
    
    @Override
    public boolean deepEquals(Expression expr) {
        if (!super.deepEquals(expr)) {
            return false;
        }
        UnionPathIterator upi = (UnionPathIterator)expr;
        if (null != this.m_exprs) {
            int n = this.m_exprs.length;
            if (null == upi.m_exprs || upi.m_exprs.length != n) {
                return false;
            }
            int i = 0;
            while (i < n) {
                if (!this.m_exprs[i].deepEquals(upi.m_exprs[i])) {
                    return false;
                }
                ++i;
            }
        }
        else if (null != upi.m_exprs) {
            return false;
        }
        return true;
    }
    
    class iterOwner implements ExpressionOwner
    {
        int m_index;
        
        iterOwner(int index) {
            this.m_index = index;
        }
        
        @Override
        public Expression getExpression() {
            return UnionPathIterator.this.m_exprs[this.m_index];
        }
        
        @Override
        public void setExpression(Expression exp) {
            if (!(exp instanceof LocPathIterator)) {
                WalkingIterator wi = new WalkingIterator(UnionPathIterator.this.getPrefixResolver());
                FilterExprWalker few = new FilterExprWalker(wi);
                wi.setFirstWalker(few);
                few.setInnerExpression(exp);
                wi.exprSetParent(UnionPathIterator.this);
                few.exprSetParent(wi);
                exp.exprSetParent(few);
                exp = wi;
            }
            else {
                exp.exprSetParent(UnionPathIterator.this);
            }
            UnionPathIterator.this.m_exprs[this.m_index] = (LocPathIterator)exp;
        }
    }
}
